/*-----------------------------------------------------------------------

  File  : labsolve.c

  Author: Stephan Schulz (schulz@eprover.org)

  Contents

  Program reading and solving simple ASCII labyrinths. New version
  using Dijksta's algorithm with rather simple data structure (node
  costs are stord in a 2D-Array, the list of open nodes is stored
  as an array of coordinate tuples.

  Copyright 2015-2018 by the author.
  This code is released under the GNU General Public Licence, version
  2, or, at your choice, any later version.

  Created:  Tue Jun  2 18:01:34 CEST 2015

  -----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <ctype.h>
#include <limits.h>
#include <assert.h>

#define MAXCOLS  2048
#define MAXROWS  2048


/* Dijkstr nodes */

typedef struct nodeinfo
{
   int x;
   int y;
}NodeInfoCell, NodeInfo_p;


/* Data type: Labyrinth - ASCII map, costs, direction */

typedef struct labyrinth
{
   char lab[MAXCOLS+2][MAXROWS];
   long costs[MAXCOLS][MAXROWS];
   bool expanded[MAXCOLS][MAXROWS];
   NodeInfoCell openlist[MAXCOLS*MAXROWS];
   int openpos;
   int maxrow;
   int maxcol;
   int startx;
   int starty;
   int treasurex;
   int treasurey;
}LabCell, *Lab_p;



/* Data structure for represeting individual moves */

typedef struct relpos
{
   int dx;
   int dy;
}RelPos;

/* Legal moves in the base case */

RelPos straight_moves[] =
{
   { 0,  1},
   { 0, -1},
   { 1,  0},
   {-1,  0},
   {INT_MAX, INT_MAX}
};

/* Legal moves if diagonals are ok */

RelPos all_moves[] =
{
   { 0,  1},
   { 0, -1},
   { 1,  0},
   {-1,  0},
   { 1,  1},
   { 1, -1},
   {-1,  1},
   {-1, -1},
   {INT_MAX, INT_MAX}
};




/*-----------------------------------------------------------------------
//
// Function: xmalloc()
//
//   Thin wrapper around malloc() - fail noisily (with error message)
//   if no memory is available.
//
// Global Variables: -
//
// Side Effects    : Only via malloc() or in case of error
//
/----------------------------------------------------------------------*/

void* xmalloc(size_t size)
{
   void *mem = malloc(size);
   if(!mem)
   {
      fprintf(stderr, "Out of memory\n");
      exit(EXIT_FAILURE);
   }
   return mem;
}

/*-----------------------------------------------------------------------
//
// Function: xstrdup()
//
//   Thin wrapper around strdup() - fail noisily (with error message)
//   if no memory is available.
//
// Global Variables: -
//
// Side Effects    : Only via malloc() or in case of error
//
/----------------------------------------------------------------------*/

char* xstrdup(char* str)
{
   void *newstr = strdup(str);
   if(!newstr)
   {
      fprintf(stderr, "Out of memory\n");
      exit(EXIT_FAILURE);
   }
   return newstr;
}

/*-----------------------------------------------------------------------
//
// Function: LabInit()
//
//   Verify and Initialize the lab: seach for and record Start and
//   Treasure positions. End with error if they are not found.
//
// Global Variables: -
//
// Side Effects    : Only in error case.
//
/----------------------------------------------------------------------*/

void LabInit(Lab_p lab)
{
   int i,j;
   bool start = false, treasure = false;

   lab->openpos = 0;
   lab->maxcol = 0;
   for(i=0; i< lab->maxrow; i++)
   {
      for(j=0; j<MAXCOLS; j++)
      {
         if((lab->lab[i][j]=='\n') && (j>lab->maxcol))
         {
            lab->maxcol = j;
         }
         lab->costs[i][j] = LONG_MAX;
         lab->expanded[i][j] = false;
         if(lab->lab[i][j] == 'S')
         {
            if(start)
            {
               fprintf(stderr, "Error: Multiple start positions\n");
               exit(EXIT_FAILURE);
            }
            lab->startx = i;
            lab->starty = j;
            start = true;
         }
         if(lab->lab[i][j] == 'X')
         {
            if(treasure)
            {
               fprintf(stderr, "Error: Multiple treasure positions\n");
               exit(EXIT_FAILURE);
            }
            lab->treasurex = i;
            lab->treasurey = j;
            treasure = true;
         }
      }
   }
   if(!start)
   {
      fprintf(stderr, "Error: No start position\n");
      exit(EXIT_FAILURE);
   }
   if(!treasure)
   {
      fprintf(stderr, "Error: No treasure position\n");
      exit(EXIT_FAILURE);
   }
}



/*-----------------------------------------------------------------------
//
// Function: LabRead()
//
//   Read a labyrinth from an input stream and return it.
//
// Global Variables: -
//
// Side Effects    : Input, memory.
//
/----------------------------------------------------------------------*/

Lab_p LabRead(FILE* in)
{
   Lab_p handle = xmalloc(sizeof(LabCell));
   char  *inp;

   handle->maxrow = 0;

   while((inp=fgets(handle->lab[handle->maxrow], MAXCOLS, in)))
   {
      handle->maxrow++;
      if(handle->maxrow>MAXROWS)
      {
         fprintf(stderr, "Labyrinth too long\n");
         exit(EXIT_FAILURE);
      }
   }
   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: LabPrint()
//
//   Print the current state of a labyrinth.
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void LabPrint(FILE* out, Lab_p lab)
{
   int i;

   for(i=0; i<lab->maxrow; i++)
   {
      fprintf(out, "%s", lab->lab[i]);
   }
}

/*-----------------------------------------------------------------------
//
// Function: LabClear()
//
//   Remove "." and "!"  markers from the Lab.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void LabClear(Lab_p lab)
{
   int i,j;

   for(i=0; i<lab->maxrow; i++)
   {
      for(j=0; j<lab->maxcol; j++)
      {
         switch(lab->lab[i][j])
         {
         case '.':
         case '!':
               lab->lab[i][j] = ' ';
               break;
         default:
               break;
         }
      }
   }
}



/*-----------------------------------------------------------------------
//
// Function: LabUpdateOpenNode()
//
//  Add a new node to the list of open nodes.
//
// Global Variables:
//
// Side Effects    :
//
/----------------------------------------------------------------------*/

void LabUpdateOpenNode(Lab_p lab, int x, int y, int cost)
{
   if(lab->costs[x][y] == LONG_MAX)
   {
      if(lab->openpos == MAXCOLS*MAXROWS)
      {
         fprintf(stderr, "Open list overflow\n");
         exit(1);
      }
      lab->openlist[lab->openpos].x = x;
      lab->openlist[lab->openpos].y = y;
      lab->openpos++;
   }
   lab->costs[x][y]=cost;
}


/*-----------------------------------------------------------------------
//
// Function: LabFindBestNode()
//
//   Find the best node in the open list.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

int LabFindBestNode(Lab_p lab)
{
   int i, best = -1;
   long bestcost = LONG_MAX;
   for(i=0; i< lab->openpos; i++)
   {
      if(lab->costs[lab->openlist[i].x][lab->openlist[i].y] < bestcost)
      {
         best = i;
         bestcost = lab->costs[lab->openlist[i].x][lab->openlist[i].y];
      }
   }
   return best;
}


/*-----------------------------------------------------------------------
//
// Function: LabRemoveOpenNode()
//
//   Remove a node from the open node list (by swapping it with the
//   last one and decrementing the openpos-counter by one).
//
// Global Variables:
//
// Side Effects    :
//
/----------------------------------------------------------------------*/

void LabRemoveOpenNode(Lab_p lab, int node)
{
   lab->openpos--;
   lab->openlist[node] = lab->openlist[lab->openpos];
}



/*-----------------------------------------------------------------------
//
// Function: LabSolve()
//
//   For all reachable fields, compute the cost of going there.
//
// Global Variables: -
//
// Side Effects    : Computes the lab costs and directions.
//
/----------------------------------------------------------------------*/

long LabSolve(Lab_p lab, RelPos moves[], int x, int y, bool print)
{
   int i;
   long res;
   char current;

   if(print)
   {
      system("clear");
      LabPrint(stdout, lab);
      printf("Exploring: %d, %d\n", x,y);
      system("sleep 0.01");
   }
   current = lab->lab[x][y];
   switch(current)
   {
   case '#': return -1;
   case '.': return -1;
   case '!': return -1;
   case 's': return -1;
   case 'X': return 0;
   default:
         break;
   }
   if(lab->lab[x][y] == ' ')
   {
      lab->lab[x][y] = '.';
   }
   for(i=0; moves[i].dx != INT_MAX; i++)
   {
      res = LabSolve(lab, moves, x+moves[i].dx, y+moves[i].dy, print);
      if(res != -1)
      {
         return res+1;
      }
   }
   lab->lab[x][y] = '!';
   return -1;
}

/*-----------------------------------------------------------------------
//
// Function: find_next_node()
//
//   Find the unexpanded field in the labyrinth with the lowest
//   estimated cost and return it (via xres/yres). Also remove it from
//   the open node list. If none is found, return false (and leave
//   xres/yres unchanged).
//
//   This is just a thin layer aroun LabFindBestNode() and
//   LabRemoveOpenNode().
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool find_next_node(Lab_p lab, int* xres, int* yres)
{
   int cand = LabFindBestNode(lab);

   if(cand == -1)
   {
      return false;
   }
   *xres = lab->openlist[cand].x;
   *yres = lab->openlist[cand].y;
   LabRemoveOpenNode(lab, cand);

   return true;
}



/*-----------------------------------------------------------------------
//
// Function: expand_node()
//
//   Expand a node - set the new estimates for all neighbouring nodes,
//   and mark it as expanded. Update the open node list.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void expand_node(Lab_p lab, RelPos moves[], int x, int y)
{
   int i, xnew, ynew;
   long cost = lab->costs[x][y], newcost;

   lab->expanded[x][y] = true;

   for(i=0; moves[i].dx != INT_MAX; i++)
   {
      newcost = LONG_MAX;
      xnew = x+moves[i].dx;
      ynew = y+moves[i].dy;
      switch(lab->lab[xnew][ynew])
      {
      case '#':
            break;
      case '.':
      case ' ':
      case 'X':
            newcost = cost+1;
            break;
      case '1':
      case '2':
      case '3':
      case '4':
      case '5':
      case '6':
      case '7':
      case '8':
      case '9':
            newcost = cost+lab->lab[xnew][ynew] - '0';
            break;
      default:
            break;
      }
      if(newcost < lab->costs[xnew][ynew])
      {
         //printf("Updating %d, %d\n", xnew, ynew);
         LabUpdateOpenNode(lab, xnew, ynew, newcost);
      }
   }
}



/*-----------------------------------------------------------------------
//
// Function: LabSolveOpt()
//
//    Use Dijkstra's algorithm to find the optimal path through the
//    labyrinth.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

long LabSolveOpt(Lab_p lab, RelPos moves[], bool print)
{
   int x, y;

   LabUpdateOpenNode(lab,lab->startx,lab->starty, 0);

   while(find_next_node(lab, &x, &y))
   {
      if(print)
      {
         if(lab->lab[x][y] == ' ')
         {
            lab->lab[x][y] = '.';
         }
         system("clear");
         LabPrint(stdout, lab);
         printf("Exploring: %d, %d\n", x,y);
         //system("sleep 0.05");
      }
      expand_node(lab, moves, x, y);
   }

   return lab->costs[lab->treasurex][lab->treasurey];
}

/*-----------------------------------------------------------------------
//
// Function: LabMarkRoute()
//
//   Given a lab with the cost information (from
//   Dijkstra/LabSolveOpt(), mark one of the best routes from start to
//   treasure.
//
// Global Variables: -
//
// Side Effects    : Changes the lab bu inserting '.' on the route
//
/----------------------------------------------------------------------*/

bool LabMarkRoute(Lab_p lab, RelPos moves[], bool print)
{
   int x,y, xnew, ynew, i;
   long min_cost;

   if(lab->costs[lab->treasurex][lab->treasurey] == LONG_MAX)
   {  // No route found!
      return false;
   }
   x = lab->treasurex;
   y = lab->treasurey;
   while(x!=lab->startx || y!=lab->starty)
   {
      if(lab->lab[x][y]==' ')
      {
         lab->lab[x][y]= '.';
      }
      if(print)
      {
         system("clear");
         LabPrint(stdout, lab);
         printf("Marking: %d, %d\n", x,y);
         system("sleep 0.01");
      }
      min_cost = LONG_MAX;
      for(i=0; moves[i].dx != INT_MAX; i++)
      {
         if(lab->costs[x+moves[i].dx][y+moves[i].dy]<min_cost)
         {
            xnew = x+moves[i].dx;
            ynew = y+moves[i].dy;
            min_cost = lab->costs[x+moves[i].dx][y+moves[i].dy];
         }
      }
      x = xnew;
      y = ynew;
   }
   return true;
}




/*-----------------------------------------------------------------------
//
// Function: main()
//
//   Do it all ;-). Options:
//   -a: Use all directions
//   -o: Use Dijkstra ("optimal")
//   -p: Print intermediate states
//
// Global Variables:
//
// Side Effects    :
//
/----------------------------------------------------------------------*/

int main(int argc, char *argv[])
{
   FILE *in = stdin;
   Lab_p lab;
   bool  optimal = false, print=false;
   long  cost;
   int   arg, args=argc,  farg;
   RelPos *moves = straight_moves;

   for(arg = 1; arg<argc; arg++)
   {
      if(strcmp(argv[arg], "-a")==0)
      {
         moves = all_moves;
         args--;
      }
      else if(strcmp(argv[arg], "-o")==0)
      {
         optimal = true;
         args--;
      }
      else if(strcmp(argv[arg], "-p")==0)
      {
         print = true;
         args--;
      }
      else
      {
         farg = arg;
      }
   }

   if(args > 2)
   {
      fprintf(stderr, "Usage: %s [-a] [-o] [-p ] [<file>]\n", argv[0]);
      exit(EXIT_FAILURE);
   }
   if(args == 2)
   {
      in = fopen(argv[farg], "r");
      if(!in)
      {
         perror(argv[0]);
         exit(EXIT_FAILURE);
      }
   }
   lab = LabRead(in);
   LabInit(lab);
   if(optimal)
   {
      cost = LabSolveOpt(lab, moves, print);
      LabClear(lab);
      if(cost != LONG_MAX)
      {
         LabMarkRoute(lab, moves, print);
      }
   }
   else
   {
      cost = LabSolve(lab, moves, lab->startx, lab->starty, print);
   }
   if(cost==-1 || cost == LONG_MAX)
   {
      //LabPrint(stdout, lab);
      fprintf(stdout, "There is no way from S to X\n");
   }
   else
   {
      LabPrint(stdout, lab);
      printf("Cost: %ld\n", cost);
   }
   exit(EXIT_SUCCESS);
}

/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/
